#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <signal.h>
#include "wios.h"
#include "util.h"
#include "alsa.h"
#include "lsf.h"
#include "../config.h"

#ifdef HAVE_SINICH
#include "tdbd.h"
#endif

#ifdef HAVE_RASP
#include "rasp.h"
#include "rasp_lcw.h"
#endif

#include "wios_util.h"


/* STATIC FUNCTION */
/* -------------------------------------------------- */

static void print_usage(const char *cmd)
{
  printf("\n");
  printf("Usage: %s [OPTION...]\n", cmd);
  printf("wios: v%s\n", WIOS_VERSION);
  printf("\n");
  printf(" Option:                      Description:\n");
  printf("\n");
  printf(" Display Option:\n");
  printf("  -h|--help                   Display version number and usage information\n");
  printf("  -v|--verbos-level LEVEL     Set verbos level (default: %d) \n", DEFAULT_WIOS_VERBOS_LEVEL);
  printf("                                0: no message\n");
  printf("                                1: file, device and result information\n");
  printf("                                2: display details of processing\n");
  printf("  -w|--warn-level LEVEL       Set warning level (default: %d)\n", DEFAULT_WIOS_WARN_LEVEL);
  printf("                                0: no warning\n");
  printf("                                1: only over-run, under-run\n");
  printf("                                2: all warning (over-run, under-run, clipping)\n");
  printf("\n");
  printf(" Mode Option:\n");
  printf("  -p|--play                   Playback D/A data file\n");
  printf("  -r|--rec                    Capture and output A/D data file\n");
  printf("  -s|--sync                   Synchronize playback and capture\n");
  printf("\n");
  printf(" File and Device Name Option:\n");
  printf("  -i|--da-file FILE           Playback this D/A file (default: %s)\n", DEFAULT_DAFILE_NAME);
  printf("  -o|--ad-file FILE           Set output A/D file name (default: %s)\n", DEFAULT_ADFILE_NAME);
  printf("  -x|--dev-type DEVICE_NUM    Specify which device to use (default: 0)\n");
  printf("                                DEVICE_NUM: (ALSA) 0\n");
  printf("                                            (TDBD) 1\n");
  printf("                                            (RASP) 2\n");
  printf("  -y|--da-dev-type DEVICE_NUM Specify the device for only DA in sync mode\n");
  printf("  -z|--ad-dev-type DEVICE_NUM Specify the device for only AD in sync mode\n");
  printf("  -d|--da-dev-name DEVICE     Use this device for D/A\n");
  printf("  -a|--ad-dev-name DEVICE     Use this device for A/D\n");
  printf("                              default: (ALSA) %s\n", ALSA_DEFAULT_DEVICE_NAME);
  printf("                                       (TDBD) %s\n", TDBD_DEFAULT_DEVICE_NAME);
  printf("                                       (RASP) %s\n", RASP_DEFAULT_DEVICE_NAME);
  printf("\n");
  printf(" Output File Format Option:\n");
  printf("  -e|--encoding ENCODING      Set encoding (default: %s)\n", DEFAULT_ENCODING);
  printf("                              ENCODING: (ALSA) u8 16 24 32\n");
  printf("                                        (TDBD)    16      \n");
  printf("                                        (RASP)    16 24   \n");
  printf("  -f|--frequency RATE         Set sampling rate [Hz] (default: %d)\n", DEFAULT_RATE);
  printf("  -c|--channels CHANNELS      Set number of channels (default: %d)\n", DEFAULT_CHANNELS);
  printf("\n");
  printf(" Other Option:\n");
  printf("  -t|--time SECOND            Process(play/rec/sync) for this time [second]\n");
  printf("  -T|--print-time             Print play/rec/sync start time\n");
  printf("\n");
  printf(" ALSA Option:\n");
  printf("  -N|--periods PERIODS        Set the number of sections in device's ring buffer (default: %d)\n", DEFAULT_PERIODS);
  printf("  -Z|--periodsize PERIODSIZE  Set the size of processing buffer size [frame] (default: %d)\n", DEFAULT_PERIODSIZE);
  printf("\n");
  printf(" TDBD Option:\n");
  printf("  -P|--print                  Display codec information configured in TDBD\n");
  printf("  -R|--reg-read ADDRESS       Read and display registar data on this address\n");
  printf("                              each address must be 0x4000-0x48FF\n");
  printf("  -W|--reg-write ADDRESS      Write specified data on this address\n");
  printf("  -D|--reg-wdata DATA         Data which is written on \"-W\" option's address\n");
  printf("                              writing data must be 0xxx format\n");
  printf("  -U|--print_usb              Display usb descriptors of TDBD\n");
  printf("  -G|--gain                   Set palyback and capture gain to device\n");
  printf("\n");
  printf(" RASP24 Option:\n");
  printf("  -M|--monoral                Set monoral mode (output diffrent data from each board)\n");
  printf("  -I|--input-gain DAGAINLIST  Set playback gain of each channel to device\n");
  printf("  -O|--output-gain ADGAINLIST Set capture gain of each channel to device\n");
  printf("                              The number of elements in GAINLIST must be equal to channels\n");
  printf("                              GAINLIST's format is \"ELEMENT1,ELEMENT2,...\". i-th element indicates i-th channel's gain\n");
  printf("  -C|--cap-gain-all GAIN_VAL  Set capture gain of all channels to same value\n");
  printf("  -B|--play-gain-all GAIN_VAL Set playback gain of all channels to same value\n");
  // FIXME setting loopback have some bug
  printf("  -L|--loopback-da DACHLIST   Set DA loopback channels (loopback to each AD channels in option \"-L\")\n");
  printf("  -A|--loopback-ad ADCHLIST   Set AD loopback channels (loopback from DA channels in DACHLIST)\n");
  printf("                              format of CHLIST is same as GAINLIST's. i-th element indicates channel number\n");
  printf("                              i-th element of DACHLIST does loopback to i-th element of ADCHLIST\n");
}

/* 指定されたファイル、デバイス、オプション等を表示 */
static void print_file_info(const wios_file_t * ft)
{
	FILE * const stream = stdout;
	char buf_sec[16];

	sprintf(buf_sec, "all");
	if (ft->signal.execinfo.sec != WIOS_ALL_SEC)
		sprintf(buf_sec, "%d", ft->signal.execinfo.sec);
	buf_sec[sizeof(buf_sec)-1] = '\0';

	fprintf(stream, "Dev:%s, Fmt:%s, Ch:%d, Rate:%d, Time:%s, P:%d, Ps:%d (%s)\n",
		ft->dev_name,
		wios_format_to_string(ft->fmtinfo.format),
		ft->fmtinfo.channels,
		ft->fmtinfo.rate,
		buf_sec,
		ft->signal.execinfo.periods,
		ft->signal.execinfo.periodsize,
		ft->file_name);
}

/* 再生(録音)結果を表示 */
static void print_file_result(const wios_file_t * ft)
{
	FILE * const stream = stdout;

	fprintf(stream, "Success:%lld[frame](%lld[byte]), Missing:%lld[frame](%lld[byte]), Clip:%d, Max:%d, Min:%d (%s)\n",
		ft->fmtinfo.format == WIOS_PCM_FORMAT_S24 ? 
			ft->success_byte_cnt/(ft->fmtinfo.channels*3) :
			ft->success_byte_cnt/(ft->fmtinfo.channels*ft->signal.byte_per_item),
		ft->success_byte_cnt,
		ft->fmtinfo.format == WIOS_PCM_FORMAT_S24 ? 
			ft->missing_byte_cnt/(ft->fmtinfo.channels*3) :
			ft->missing_byte_cnt/(ft->fmtinfo.channels*ft->signal.byte_per_item),
		ft->missing_byte_cnt,
		ft->clips,
		ft->max_item_val,
		ft->min_item_val,
		ft->file_name);
}

/* SIGINT用シグナルハンドラ */
static void sig_int(int sig) {
	puts("");
	/* 再生(録音)ループを抜けさせる */
	loop_break_flag = WIOS_TRUE;
}

static void sig_int_init(void) {
	if (SIG_ERR == signal(SIGINT, sig_int)) {
		puts("failed to set signal handler");
		exit(1);
	}
}

static struct option const longopts[] = {
	{"help"        , no_argument      , NULL, 'h'},

	{"play"        , no_argument      , NULL, 'p'},
	{"rec"         , no_argument      , NULL, 'r'},
	{"sync"        , no_argument      , NULL, 's'},

	{"da-file"     , required_argument, NULL, 'i'},
	{"ad-file"     , required_argument, NULL, 'o'},
	{"dev-type"    , required_argument, NULL, 'x'},
	{"da-dev-type" , required_argument, NULL, 'y'},
	{"ad-dev-type" , required_argument, NULL, 'z'},
	{"da-dev-name" , required_argument, NULL, 'd'},
	{"ad-dev-name" , required_argument, NULL, 'a'},
	{"encoding"    , required_argument, NULL, 'e'},
	{"frequency"   , required_argument, NULL, 'f'},
	{"channels"    , required_argument, NULL, 'c'},
	{"time"        , required_argument, NULL, 't'},
	{"print-time"  , no_argument,       NULL, 'T'},

	{"warn-level"  , required_argument, NULL, 'w'},
	{"verbos-level", required_argument, NULL, 'v'},

	{"periods"     , required_argument, NULL, 'N'},
	{"periodsize"  , required_argument, NULL, 'Z'},

	{"gain"        , required_argument, NULL, 'G'},
	{"print"       , no_argument      , NULL, 'P'},
	{"reg-read"    , required_argument, NULL, 'R'},
	{"reg-write"   , required_argument, NULL, 'W'},
	{"reg-wdata"   , required_argument, NULL, 'D'},
	{"print-usb"   , no_argument      , NULL, 'U'},

	{"monoral"     , no_argument      , NULL, 'M'},
	{"input-gain"  , required_argument, NULL, 'I'},
	{"output-gain" , required_argument, NULL, 'O'},
	{"loopback-da" , required_argument, NULL, 'L'},
	{"loopback-ad" , required_argument, NULL, 'A'},
	{"cap-gain-all", required_argument, NULL, 'C'},
	{"play-gain-all",required_argument, NULL, 'B'},

	{NULL, 0, NULL, 0}
};

static const char * optstr = "+A:a:B:c:C:d:D:e:f:G:hi:I:L:MN:o:O:pPrR:st:TUv:w:W:x:y:z:Z:";

static int wios_getopt_long(int argc, char ** argv)
{
	int c;
	int digit_optind = 0;
	int dev_name_specified = 0;

	char * reg_wr_addr=NULL, * reg_wr_data=NULL;
	
	while (1) {
		int this_option_optind = optind ? optind : 1;
		int i, n, option_index = 0;

		c = getopt_long(argc, argv, optstr, longopts, &option_index);
		if (c == -1)
			break;

		switch (c) {
			case 'h':
				print_usage(myname);
				exit(1);

			case 'v':
				n = atoi(optarg);
				if (n > WIOS_VERBOS_LEVEL_MAX) n = WIOS_VERBOS_LEVEL_MAX;
				if (n < WIOS_VERBOS_LEVEL_MIN) n = WIOS_VERBOS_LEVEL_MIN;
				if (n <= 0) print_status_flag = print_result_flag = WIOS_FALSE;
				if (n >= 2) print_mesg_flag = WIOS_TRUE;
				wios_verbos_level = n;
				break;

			case 'w':
				n = atoi(optarg);
				if (n > WIOS_WARN_LEVEL_MAX) n = WIOS_WARN_LEVEL_MAX;
				if (n < WIOS_WARN_LEVEL_MIN) n = WIOS_WARN_LEVEL_MIN;
				wios_warn_level = n;
				break;

			case 'p': wios_mode = wios_play; break;
			case 'r': wios_mode = wios_rec;  break;
			case 's': wios_mode = wios_sync; break;

			case 'i': da_path = optarg; break;
			case 'o': ad_path = optarg; break;

			/* -d と -a は，とりあえず先に指定された方に揃える */
			case 'd':
				da_dev  = optarg;
				if (0==dev_name_specified) {
					ad_dev = da_dev; // 先にda_devが指定されたので，ad_devをad_devに
					dev_name_specified = 1;
					use_da_dev_option_flag = WIOS_TRUE;
				}
				break;
			case 'a':
				ad_dev  = optarg;
				if (0==dev_name_specified) {
					da_dev = ad_dev; // 先にad_devが指定されたので，da_devをad_devに
					dev_name_specified = 1;
					use_ad_dev_option_flag = WIOS_TRUE;
				}
				break;

			case 'x':
				wios_dev = atoi(optarg);
				break;

			case 'y':
				use_2dev_option_flag = WIOS_TRUE;
				wios_da_dev = atoi(optarg);
				break;

			case 'z':
				use_2dev_option_flag = WIOS_TRUE;
				wios_ad_dev = atoi(optarg);
				break;

			case 'e':
				for (i = 0; i < ARRAY_SIZE(supported_encoding); i++)
					if (!strncmp(optarg, supported_encoding[i], 2)) break;
				if (i == ARRAY_SIZE(supported_encoding)) {
					fprintf(stderr, "Encoding `%s' is not supported. ", optarg);
					fprintf(stderr, "Supported encoding: ");
					for (i = 0; i < ARRAY_SIZE(supported_encoding); i++)
						fprintf(stderr, "%s ", supported_encoding[i]);
					fprintf(stderr, "\n");
					exit(1);
				}
				ad_fmtinfo.format = wios_format[i];
				use_enc_option_flag = WIOS_TRUE;
				break;

			case 'f':
				if ((n = atoi(optarg)) <= 0) {
					wios_fail("Rate `%s' is not a positive number", optarg);
					exit(1);
				}
				ad_fmtinfo.rate = n;
				break;

			case 'c':
				if ((n = atoi(optarg)) <= 0) {
					wios_fail("Channels `%s' is not a positive number", optarg);
					exit(1);
				}
				ad_fmtinfo.channels = n;
				use_ch_option_flag  = WIOS_TRUE;
				break;

			case 't':
				if ((n = atoi(optarg)) < WIOS_MIN_SEC) {
					wios_fail("Time `%s' must be %d < time", optarg, WIOS_MIN_SEC);
					exit(1);
				}
				execinfo.sec = n;
				break;

      case 'T':
        print_time_flag = WIOS_TRUE;
        break;

			case 'N':
				if ((n = atoi(optarg)) <= 0) {
					wios_fail("Periods `%s' is not a positive number", optarg);
					exit(1);
				}
				execinfo.periods = n;
				break;

			case 'Z':
				if ((n = atoi(optarg)) <= 0) {
					wios_fail("Periodsize `%s' is not a positive number", optarg);
					exit(1);
				}
				execinfo.periodsize = n;
				break;

#ifdef HAVE_SINICH
			case 'P':
				print_codec_flag = WIOS_TRUE;
				break;

			case 'R':
				tdbd_register_read(optarg);
				exit(1);

			case 'W':
				reg_wr_addr = optarg;
				if (reg_wr_data) {
					tdbd_register_write(reg_wr_addr, reg_wr_data);
					exit(1);
				}
				break;
				
			case 'D':
				reg_wr_data = optarg;
				if (reg_wr_addr) {
					tdbd_register_write(reg_wr_addr, reg_wr_data);
					exit(1);
				}
				break;

			case 'U':
				print_descr_flag = WIOS_TRUE;
				break;

			case 'G':
				if ((n = atoi(optarg)) <= 0) {
					wios_fail("Gain `%s' is not a positive number", optarg);
					exit(1);
				}
				set_gain_val = n;
				break;
#endif

#ifdef HAVE_RASP
			case 'M':
				da_monoral_flag = WIOS_TRUE;
				break;

			case 'I': rasp_ad_gain_str  = optarg; break;
			case 'O': rasp_da_gain_str  = optarg; break;
			case 'L': rasp_da_lb_ch_str = optarg; break;
			case 'A': rasp_ad_lb_ch_str = optarg; break;
			case 'C': rasp_ad_gain_val  = atoi(optarg); break;
			case 'B': rasp_da_gain_val  = atoi(optarg); break;
#endif

			default:
				/* printf("?? getopt returned character code 0%o ??\n", c); */
				wios_fail("end");
				exit(1);
		}
	}

	if (reg_wr_addr || reg_wr_data) exit(1);
	
	return WIOS_SUCCESS;

error:
	exit(1);
}

/* 再生用ファイルを開いてデバイスを初期化する */
static int prepare_playback(void)
{
	if (print_mesg_flag) {
		puts("Start preparation for playback");
		puts("Start DA file open");
		printf("D/A device: %s\n", da_dev);
	}

	if (!use_2dev_option_flag) {
		if ((ftda = handler->open_read(da_path, da_dev, &execinfo)) == NULL) exit(1);
	} else {
		if ((ftda = da_handler->open_read(da_path, da_dev, &execinfo)) == NULL) exit(1);
	}

	if (print_mesg_flag) {
		puts("End DA file open");
		puts("Start device init for playback");
	}

	if (!use_2dev_option_flag) {
		if (handler->device_init(ftda) == WIOS_EOF) exit(1);
	} else {
		if (da_handler->device_init(ftda) == WIOS_EOF) exit(1);
	}

	if (print_mesg_flag) {
		puts("End device init for playback");
		puts("End preparation for playback");
	}

	if (print_status_flag)
		print_file_info(ftda);
	
	return WIOS_SUCCESS;
}

/* 録音用ファイルを開いてデバイスを初期化する */
static int prepare_capture(void)
{
	if (print_mesg_flag) {
		puts("Start preparation for capture");
		puts("Start AD file open");
		printf("A/D device: %s\n", ad_dev);
	}

	if (wios_mode == wios_sync) {
		int channels = ad_fmtinfo.channels;
		int format   = ad_fmtinfo.format;

		/* ad_fmtinfoにftdaのフォーマットをセット */
		wios_set_fmtinfo(ftda, &ad_fmtinfo);

		/* チャンネル数が指定されていたら設定 */
		if (use_ch_option_flag) ad_fmtinfo.channels = channels;
		/* フォーマットが指定されていたら設定 */
		if (use_enc_option_flag) ad_fmtinfo.format  = format;
	}

	/* RASP-LCW : ４チャンネル/２４ビット固定 */
	if (wios_dev == wios_rasp_lcw) {
		ad_fmtinfo.channels = 4;
		ad_fmtinfo.format  = WIOS_PCM_FORMAT_S24;
	}

	if (!use_2dev_option_flag) {
		if ((ftad = handler->open_write(ad_path, ad_dev, &execinfo, &ad_fmtinfo)) == NULL)
			exit(1);
	} else {
		if ((ftad = ad_handler->open_write(ad_path, ad_dev, &execinfo, &ad_fmtinfo)) == NULL)
			exit(1);
	}

	if (print_mesg_flag) {
		puts("End AD file open");
		puts("Start device init for capture");
	}

	if (!use_2dev_option_flag) {
		if (handler->device_init(ftad) == WIOS_EOF) exit(1);
	} else {
		if (ad_handler->device_init(ftad) == WIOS_EOF) exit(1);
	}

	if (print_mesg_flag) {
		puts("End device init for capture");
		puts("End preparation for capture");
	}

	if (print_status_flag)
		print_file_info(ftad);
	
	return WIOS_SUCCESS;
}

/* 再録同期の準備をした後同期を開始させる */
static int start_sync(void)
{
	if (print_mesg_flag)
		puts("Start preparation for synchronizing");

	/* 再生と録音用スレッドを準備 */
	pthread_mutex_lock(&ready_mutex);

	if (!use_2dev_option_flag)
		pthread_create(&da_thread, 0, handler->playback, (void *)ftda);
	else
		pthread_create(&da_thread, 0, da_handler->playback, (void *)ftda);
	pthread_cond_wait(&ready_cond, &ready_mutex);

	if (!use_2dev_option_flag)
		pthread_create(&ad_thread, 0, handler->capture, (void *)ftad);
	else
		pthread_create(&ad_thread, 0, ad_handler->capture, (void *)ftad);
	pthread_cond_wait(&ready_cond, &ready_mutex);

	pthread_mutex_unlock(&ready_mutex);

	sync_start_flag = WIOS_TRUE;

	if (print_mesg_flag) {
		puts("End preparation for synchronizing");
		puts("Start synchronizing");
	}

	/* 再録同期開始 */
	pthread_cond_broadcast(&sync_cond);

	pthread_join(ad_thread, NULL);
	pthread_join(da_thread, NULL);

	if (print_mesg_flag)
		puts("End synchronizing");

	return WIOS_SUCCESS;
}

static void wios_init(void)
{
	wios_mode = wios_play;

	myname = "wios";

	pthread_mutex_init(&ready_mutex, NULL);
	pthread_mutex_init(&sync_mutex, NULL);
	pthread_cond_init(&ready_cond, NULL);
	pthread_cond_init(&sync_cond, NULL);

	ftda = NULL;
	ftad = NULL;

	handler = NULL;
	da_handler = NULL;
	ad_handler = NULL;

	wios_dev = wios_unknown;
	wios_da_dev = wios_unknown;
	wios_ad_dev = wios_unknown;

	wios_warn_level = DEFAULT_WIOS_WARN_LEVEL;
	wios_verbos_level = DEFAULT_WIOS_VERBOS_LEVEL;

	ad_path = DEFAULT_ADFILE_NAME;
	da_path = DEFAULT_DAFILE_NAME;

	ad_dev  = ALSA_DEFAULT_DEVICE_NAME;
	da_dev  = ALSA_DEFAULT_DEVICE_NAME;

	sync_start_flag = WIOS_FALSE;

	playback_end_flag = WIOS_FALSE;
	loop_break_flag   = WIOS_FALSE;

	print_result_flag = WIOS_TRUE;
	print_status_flag = WIOS_TRUE;
	print_mesg_flag   = WIOS_FALSE;
  print_time_flag   = WIOS_FALSE;

	use_ch_option_flag  = WIOS_FALSE;
	use_enc_option_flag = WIOS_FALSE;
	use_da_dev_option_flag = WIOS_FALSE;
	use_ad_dev_option_flag = WIOS_FALSE;
	use_2dev_option_flag = WIOS_FALSE;

	wios_execinfo_init(&execinfo);

	wios_fmtinfo_init(&ad_fmtinfo);

	wios_mode = wios_help;
	wios_dev = wios_unknown;

#ifdef HAVE_SINICH
	set_gain_val = 0;

	print_codec_flag = WIOS_FALSE;
	print_descr_flag = WIOS_FALSE;
#endif

#ifdef HAVE_RASP
	da_monoral_flag = WIOS_FALSE;
	lb_flag = WIOS_FALSE;

	rasp_ad_gain_str  = NULL;
	rasp_da_gain_str  = NULL;
	rasp_ad_lb_ch_str = NULL;
	rasp_da_lb_ch_str = NULL;
#endif
}

#if 0
/* 接続デバイスの自動判別 */
/* 不安定なので使わない */
static wios_device_t wios_device_check(void) {
	da_dev="/dev/sinichusb0";
	ad_dev="/dev/sinichusb0";
	if (wios_mode == wios_rec) {
		if (tdbd_device_check(ad_dev)) return wios_tdbd;
		return wios_alsa;
	} else {
		if (tdbd_device_check(da_dev)) return wios_tdbd;
		return wios_alsa;
	}
}
#endif

/* 指定されたデバイスオプションの正しさをチェック */
void wios_check() {
	if (wios_mode != wios_sync) {
		if (wios_dev == wios_unknown || use_2dev_option_flag) {
			print_usage(myname);
			exit(1);
		}

	} else {
		if (!use_2dev_option_flag) {
			if (wios_dev == wios_unknown) {
				print_usage(myname);
				exit(1);
			}
		} else {
			if ((wios_dev != wios_unknown) || 
				  (wios_da_dev == wios_unknown || wios_ad_dev == wios_unknown)) {
				print_usage(myname);
				exit(1);
			}
		}
	}
}

/* オプションに対応するデバイスハンドラをセット */
void wios_set_handler() {
	if (wios_dev == wios_alsa) {
		handler = wios_set_alsa_format_handler();
	}

	if (wios_da_dev == wios_alsa) {
		da_handler = wios_set_alsa_format_handler();
	}
	if (wios_ad_dev == wios_alsa) {
		ad_handler = wios_set_alsa_format_handler();
	}

#ifdef HAVE_SINICH
	if (wios_dev == wios_tdbd) {
		handler = wios_set_tdbd_format_handler();
		if (!use_da_dev_option_flag) da_dev = TDBD_DEFAULT_DEVICE_NAME;
		if (!use_ad_dev_option_flag) ad_dev = TDBD_DEFAULT_DEVICE_NAME;
	}

	if (wios_da_dev == wios_tdbd) {
		da_handler = wios_set_tdbd_format_handler();
		if (!use_da_dev_option_flag) da_dev = TDBD_DEFAULT_DEVICE_NAME;
	}
	if (wios_ad_dev == wios_tdbd) {
		ad_handler = wios_set_tdbd_format_handler();
		if (!use_ad_dev_option_flag) ad_dev = TDBD_DEFAULT_DEVICE_NAME;
	}
#endif

#ifdef HAVE_RASP
	if (wios_dev == wios_rasp) {
		handler = wios_set_rasp_format_handler();
		if (!use_da_dev_option_flag) da_dev = RASP_DEFAULT_DEVICE_NAME;
		if (!use_ad_dev_option_flag) ad_dev = RASP_DEFAULT_DEVICE_NAME;
	}

	if (wios_da_dev == wios_rasp) {
		da_handler = wios_set_rasp_format_handler();
		if (!use_da_dev_option_flag) da_dev = RASP_DEFAULT_DEVICE_NAME;
	}
	if (wios_ad_dev == wios_rasp) {
		ad_handler = wios_set_rasp_format_handler();
		if (!use_ad_dev_option_flag) ad_dev = RASP_DEFAULT_DEVICE_NAME;
	}

	if (wios_dev == wios_rasp_lcw) {
		handler = wios_set_rasp_lcw_format_handler();
	}

#endif

  if ((handler && !da_handler && !ad_handler) ||
      (!handler && da_handler && ad_handler)) {
    ;
  } else {
		print_usage(myname);
		puts("hoge");
		exit(1);
  }
}



/* MAIN FUNCTION */
/* -------------------------------------------------- */

int main(int argc, char ** argv)
{
	wios_init();
	sig_int_init();
	wios_getopt_long(argc, argv);

	wios_check();

	wios_set_handler();

	switch (wios_mode) {

		case wios_play: /* 再生のみ */
			if (prepare_playback() == WIOS_EOF)
				exit(1);
			handler->playback(ftda);
			if (print_result_flag)
				print_file_result(ftda);
			handler->finish(ftda);
			break;

		case wios_rec: /* 録音のみ */
			if (prepare_capture() == WIOS_EOF)
				exit(1);
			handler->capture(ftad);
			if (print_result_flag)
				print_file_result(ftad);
			handler->finish(ftad);
			break;

		case wios_sync: /* 再録同期 */
			if (prepare_playback() == WIOS_EOF)
				exit(1);
			if (prepare_capture() == WIOS_EOF)
				exit(1);
			start_sync();
			if (print_result_flag) {
				print_file_result(ftda);
				print_file_result(ftad);
			}
			if (!use_2dev_option_flag) {
				handler->finish(ftda);
				handler->finish(ftad);
			} else {
				da_handler->finish(ftda);
				ad_handler->finish(ftad);
			}
			break;

		default:
			print_usage(myname);
			break;
	}

	puts("wios end");

	return 0;
}
